@@ -5,12 +5,23 @@ the diseases they do or do not have) by constructing and analysing hypergraph
 data structures.
 """
 
-import itertools
-import numpy as np
-import numba
-import scipy.sparse as ssp
-import scipy.sparse.linalg as sspl
-import scipy.stats as sst
+from itertools import combinations, chain
+
+# numpy imports
+# submodules
+from numpy import random
+from numpy.linalg import norm
+# types
+from numpy import int8, int32, uint8, int64, float64
+# array tools
+from numpy import zeros, zeros_like, ones_like, array, arange
+# functions
+from numpy import ceil, log, sqrt, multiply, abs, unique, where
+
+import numba # imported types would clash with some of the numpy imports.
+
+from scipy.sparse import lil_matrix, csr_matrix, diags
+from scipy.sparse.linalg import eigsh
 
 from time import time
 
@@ -18,99 +29,6 @@ from time import time
 ##########################################
 ## Numba compiled functions here
 
-@numba.jit(
-    nopython=True,
-    nogil=True,
-    fastmath=True,
-)
-def _binomial_rvs(N, p):
-    """
-    Generates a single random integer from a binomial distribution given
-    a population size and a probability.
-    """
-
-    count = 0
-    while True:
-        wait = np.ceil(np.log(np.random.rand()) / np.log(1-p))
-        if wait > N:
-            return count
-        count += 1
-        N -= wait
-
-
-@numba.jit(
-    nopython=True,
-    nogil=True,
-    parallel=True,
-    fastmath=True,
-)
-def randomize_weights(N_array, p_array):
-    """
-    Generates a set of binomiallly distributed random weights for use in the
-    perturbation uncertainty calculation method. Given an array of populations
-    and probabilities, this function returns the equivalent of
-
-    [sst.binom(N, p).rvs(samples) / N for N, p in zip(N_array, p_array)]
-
-    Parameters
-    ----------
-
-        N_array : numpy.array
-            An array of population sizes
-
-        p_array : numpy.array
-            An array of probabilities
-
-    Returns
-    -------
-
-        numpy.array : an array of random numbers of length len(N_array)
-    """
-    out = np.zeros(len(N_array), dtype=np.float64)
-    for i in numba.prange(len(N_array)):
-        out[i] = _binomial_rvs(N_array[i], p_array[i]) / N_array[i]
-    return out
-
-
-@numba.jit(
-    nopython=True,
-    nogil=True,
-    fastmath=True,
-)
-def _wilson_interval(num, denom):
-    """
-    Returns an estimate of the variance of the overlap coefficient (and other
-    quantities that feature the some number of people divided by the total population
-    size) derived from the Wilson score interval.
-
-    Parameters
-    ----------
-
-        num : numeric
-            the number of "successes", i.e. the numerator of the rate formula we want
-            to calculate the variance for.
-
-        denom  : numeric
-            the population size of the rate formula we want to calculate the
-            variance for.
-
-    Returns
-    -------
-
-        numpy.float64 : the estimate of the variance.
-    """
-
-    z = 1.959963984540054 # this is the 2.5th/97.5th percentile of the standard normal
-    wilson_centre = (num + 0.5 * z ** 2) / (denom + z ** 2)
-    wilson_offset = z * np.sqrt(num * (denom - num) / denom + z ** 2 / 4) / (denom + z ** 2)
-
-    return (wilson_centre - wilson_offset, wilson_centre + wilson_offset)
-    # NOTE
-    # ((x + a) - (x - a)) / 2
-    # = (2a) / 2
-    # = a
-
-
 @numba.jit(
     nopython=True,
     nogil=True,
@@ -144,32 +62,39 @@ def _overlap_coefficient(data, inds, *args):
 
         numpy.float64 : The calculated overlap coefficient.
 
-        numpy.float64 : The calculated confidence interval in the overlap coefficient.
-
         numpy.float64 : The denominator used to calculate the overlap coefficient.
 
     """
     n_diseases = inds.shape[0]
-    numerator = 0.0
-    denominator = np.zeros(shape=(n_diseases))
+    
+    if n_diseases == 1:
+        # if there is one disease, return the prevalence
+        numerator = 0.0
+        for row in range(data.shape[0]):
+            if data[row, inds[0]] > 0:
+                numerator += 1.0
+        denom = float64(data.shape[0])
+       
+    else:    
+        # if there is more than one disease, return the overlap coeffcient.
+        numerator = 0.0
+        denominator = zeros(shape=(n_diseases))
 
-    for ii in range(data.shape[0]):
-        loop_sum = 0
-        for jj in range(n_diseases):
-            loop_sum += data[ii, inds[jj]]
-            denominator[jj] += data[ii, inds[jj]]
+        for ii in range(data.shape[0]):
+            loop_sum = 0
+            for jj in range(n_diseases):
+                loop_sum += data[ii, inds[jj]]
+                denominator[jj] += data[ii, inds[jj]]
 
-        if loop_sum == n_diseases:
-            numerator += 1.0
+            if loop_sum == n_diseases:
+                numerator += 1.0
 
-    denom = denominator[0]
-    for jj in range(1, n_diseases):
-        if denominator[jj] < denom:
-            denom = denominator[jj]
+        denom = denominator[0]
+        for jj in range(1, n_diseases):
+            if denominator[jj] < denom:
+                denom = denominator[jj]
 
-    wilson_ci = _wilson_interval(numerator, denom)
-
-    return numerator / denom, wilson_ci, denom
+    return numerator / denom, denom
 
 @numba.jit(
     nopython=True,
@@ -224,10 +149,6 @@ def _compute_weights(
             The edge weight vector, a one dimensional array of length n_edges
             that contains the calculated edge weights.
 
-        numpy.array
-            The edge weight confidence interval vector, a one dimensional array of
-            length n_edges that contains the calculated variances in the edge weights.
-
         numpy.array
             The edge weight population vector, a one dimensional array of length n_edges
             that contains the population sizes used in the calculation of the the edge
@@ -237,10 +158,6 @@ def _compute_weights(
             The node weight vector, a one dimensional array of length n_nodes
             that contains the calculated node weights.
 
-        numpy.array
-            The node weight confidence interval vector, a one dimensional array of
-            length n_nodes that contains the calculated variances in the node weights.
-
         numpy.array
             The node weight population vector, a one dimensional array of length n_nodes
             that contains the population sizes used in the calculation of the the node
@@ -248,10 +165,10 @@ def _compute_weights(
     """
 
 
-    incidence_matrix = np.zeros(shape=(work_list.shape[0], data.shape[1]), dtype=np.uint8)
-    edge_weight = np.zeros(shape=work_list.shape[0], dtype=np.float64)
-    edge_weight_ci = np.zeros(shape=(work_list.shape[0], 2), dtype=np.float64)
-    edge_weight_population = np.zeros(shape=work_list.shape[0], dtype=np.float64)
+    incidence_matrix = zeros(shape=(work_list.shape[0], data.shape[1]), dtype=uint8)
+    edge_weight = zeros(shape=work_list.shape[0], dtype=float64)
+    edge_weight_ci = zeros(shape=(work_list.shape[0], 2), dtype=float64)
+    edge_weight_population = zeros(shape=work_list.shape[0], dtype=float64)
 
     # NOTE (Jim 14/6/2021)
     # There is an absolutely bizarre bug in Numba 0.53.1
@@ -275,37 +192,27 @@ def _compute_weights(
         # This call to weight_function allows users to define their
         # own weight functions with or without the optional args.
         # :-)
-        weight, ci, denom = weight_function(data, inds, *args)
+        weight, denom = weight_function(data, inds, *args)
 
         edge_weight[index] = weight
-        edge_weight_ci[index, :] = ci
         edge_weight_population[index] = denom
         for jj in range(n_diseases):
             incidence_matrix[index, inds[jj]] = 1
 
-    node_weight = np.zeros(shape=data.shape[1], dtype=np.float64)
-    node_weight_ci = np.zeros(shape=(data.shape[1], 2), dtype=np.float64)
-    node_weight_population = np.zeros(shape=data.shape[1] , dtype=np.float64)
+    node_weight = zeros(shape=data.shape[1], dtype=float64)
+    node_weight_ci = zeros(shape=(data.shape[1], 2), dtype=float64)
+    node_weight_population = zeros(shape=data.shape[1] , dtype=float64)
 
     for index in numba.prange(data.shape[1]):
-        numerator = 0.0
-        for row in range(data.shape[0]):
-            if data[row, index] > 0:
-                numerator += 1.0
-        node_weight[index] = (numerator / np.float64(data.shape[0]))
-        node_weight_ci[index, :] = _wilson_interval(
-            numerator,
-            np.float64(data.shape[0])
-        )
-        node_weight_population[index] = np.float64(data.shape[0])
+        weight, denom = weight_function(data, array([index]), *args)
+        node_weight[index] = weight
+        node_weight_population[index] = denom
 
     return (
         incidence_matrix,
         edge_weight,
-        edge_weight_ci,
         edge_weight_population,
         node_weight,
-        node_weight_ci,
         node_weight_population,
     )
 
@@ -342,17 +249,17 @@ def _bipartite_eigenvector(incidence_matrix, weight):
             The calculated uncertainties in the eigenvector elements
 
     '''
-    weighted_matrix = np.multiply(incidence_matrix, weight.reshape((-1, 1)))
+    weighted_matrix = multiply(incidence_matrix, weight.reshape((-1, 1)))
     n_edges, n_nodes = weighted_matrix.shape
     total_elems = n_edges + n_nodes
 
-    adjacency_matrix = ssp.lil_matrix((total_elems, total_elems))
+    adjacency_matrix = lil_matrix((total_elems, total_elems))
     adjacency_matrix[n_nodes:total_elems, 0:n_nodes] = weighted_matrix
     adjacency_matrix[0:n_nodes, n_nodes:total_elems] = weighted_matrix.T
 
-    eig_val, eig_vec = sspl.eigsh(adjacency_matrix, k=1)
+    eig_val, eig_vec = eigsh(adjacency_matrix, k=1)
 
-    return np.abs(eig_val[0]), np.abs(eig_vec.reshape(-1))
+    return abs(eig_val[0]), abs(eig_vec.reshape(-1))
 
 @numba.jit(
     [numba.float64[:](numba.uint8[:, :], numba.float64[:], numba.float64[:]),
@@ -395,17 +302,17 @@ def _iterate_vector(incidence_matrix, weight, vector):
     '''
 
     # we are calculating [M W M^T - diag(M W M^T)] v
-    term_1 = np.zeros_like(vector)
+    term_1 = zeros_like(vector)
 
     # 1) W M^T
-    weighted_incidence = np.zeros_like(incidence_matrix, dtype=np.float64)
+    weighted_incidence = zeros_like(incidence_matrix, dtype=float64)
     for i in numba.prange(weighted_incidence.shape[0]):
         for j in range(weighted_incidence.shape[1]):
             weighted_incidence[i, j] += incidence_matrix[i, j] * weight[j]
 
 
     # 2) W M^T v
-    intermediate = np.zeros(weighted_incidence.shape[1], dtype=vector.dtype)
+    intermediate = zeros(weighted_incidence.shape[1], dtype=vector.dtype)
     for k in numba.prange(weighted_incidence.shape[1]):
         for j in range(len(vector)):
             intermediate[k] += weighted_incidence[j, k] * vector[j]
@@ -416,13 +323,13 @@ def _iterate_vector(incidence_matrix, weight, vector):
             term_1[i] += incidence_matrix[i, k] * intermediate[k]
 
     # 4) diag(M W M^T v) can be done in one step using W M^T from before
-    subt = np.zeros_like(vector)
+    subt = zeros_like(vector)
     for i in numba.prange(len(vector)):
         for k in range(weighted_incidence.shape[1]):
             subt[i] += incidence_matrix[i, k] * weighted_incidence[i, k] * vector[i]
 
     # 5) subtract one from the other.
-    result = np.zeros_like(vector)
+    result = zeros_like(vector)
     for i in numba.prange(len(vector)):
         result[i] = term_1[i] - subt[i]
 
@@ -466,8 +373,8 @@ def _reduced_powerset(iterable, min_set=0, max_set=None):
     if max_set is None:
         max_set = len(iterable)+1
 
-    return itertools.chain.from_iterable(
-        itertools.combinations(iterable, r) for r in range(min_set, max_set)
+    return chain.from_iterable(
+        combinations(iterable, r) for r in range(min_set, max_set)
     )
 
 
@@ -496,8 +403,6 @@ class Hypergraph(object):
         self.incidence_matrix = None
         self.edge_weights = None
         self.node_weights = None
-        self.edge_weights_ci = None
-        self.node_weights_ci = None
         self.edge_weights_pop = None
         self.node_weights_pop = None
         self.edge_list = None
@@ -555,18 +460,10 @@ class Hypergraph(object):
                 The edge weight vector, a one dimensional array of length n_edges
                 that contains the calculated edge weights.
 
-            edge_weights_ci : numpy.array
-                The edge weight confidence interval vector, a one dimensional array of
-                length n_edges that contains the calculated variances in the edge weights.
-
             node_weights : numpy.array
                 The node weight vector, a one dimensional array of length n_nodes
                 that contains the calculated node weights.
 
-            node_weights_ci : numpy.array
-                The node weight confidence interval vector, a one dimensional array of
-                length n_nodes that contains the calculated variances in the node weights.
-
             edge_list : list
                 The edge list, a list of edges each element of which is a tuple of strings
                 derived from the names of the pandas.DataFrame columns and indicates the
@@ -584,7 +481,7 @@ class Hypergraph(object):
         """
 
         # construct a matrix of flags from the pandas input
-        data_array = data.to_numpy().astype(np.uint8)
+        data_array = data.to_numpy().astype(uint8)
 
         # create a list of edges (numpy array of indices)
         node_list_string = data.keys().tolist()
@@ -592,14 +489,14 @@ class Hypergraph(object):
         n_diseases = len(node_list)
 
         t = time()
-        m_data = np.unique(data_array, axis=0)
+        m_data = unique(data_array, axis=0)
 
         # I'm very grateful to Alex Lee for providing the initial version of this
         # edge pruning solution. Thank you!
 
         # Alex's solution (fails some tests because it's
         # returning a set of all diseases.
-        #unique_co_occurences = [list(np.where(elem)[0]) for elem in m_data[np.sum(m_data, axis=1)>=2]]
+        #unique_co_occurences = [list(where(elem)[0]) for elem in m_data[sum(m_data, axis=1)>=2]]
         #valid_powersets = [
         #    set(itertools.chain(
         #        *[list(itertools.combinations(elem, i)) for i in range(2, len(elem) + 1)]
@@ -608,13 +505,13 @@ class Hypergraph(object):
         #edge_list = [list(elem) for elem in set.union(*valid_powersets)]
 
         # my solution
-        m_data = m_data[np.sum(m_data, axis=1) >= 2]
+        m_data = m_data[m_data.sum(axis=1) >= 2]
         edge_list = list(set().union(*[
             list(
                 _reduced_powerset(
-                    np.where(i)[0],
+                    where(i)[0],
                     min_set=2,
-                    max_set=np.min([np.sum(i)+1, n_diseases]).astype(np.int64)
+                    max_set=array([i.sum()+1, n_diseases]).min().astype(int64)
                 )
             ) for i in m_data
         ]))
@@ -628,25 +525,23 @@ class Hypergraph(object):
         #    )
         #)
 
-        max_edge = np.max([len(ii) for ii in edge_list])
-        work_list = np.array(
+        max_edge = array([len(ii) for ii in edge_list]).max()
+        work_list = array(
             [list(ii) + [-1] * (max_edge - len(ii)) for ii in edge_list],
-            dtype=np.int8
+            dtype=int8
         )
 
         # shuffle the work list to improve runtime
-        reindex = np.arange(work_list.shape[0])
-        np.random.shuffle(reindex)
+        reindex = arange(work_list.shape[0])
+        random.shuffle(reindex)
         work_list = work_list[reindex, :]
-        edge_list = np.array(edge_list, dtype="object")[reindex].tolist()
+        edge_list = array(edge_list, dtype="object")[reindex].tolist()
 
         # compute the weights
         (inc_mat_original,
         edge_weight,
-        edge_weight_ci,
         edge_weight_pop,
         node_weight,
-        node_weight_ci,
         node_weight_pop) = _compute_weights(
             data_array,
             work_list,
@@ -661,17 +556,16 @@ class Hypergraph(object):
         inds = edge_weight > 0
         inc_mat_original = inc_mat_original[inds, :]
         edge_weight = edge_weight[inds]
-        edge_weight_ci = edge_weight_ci[inds]
-        edge_list_out = np.array(edge_list_out, dtype="object")[inds].tolist()
+        edge_weight_pop = edge_weight_pop[inds]
+        
+        edge_list_out = array(edge_list_out, dtype="object")[inds].tolist()
         # traverse the edge list one final time to make sure the edges are tuples
         edge_list_out = [tuple(ii) for ii in edge_list_out]
 
         self.incidence_matrix = inc_mat_original
         self.edge_weights = edge_weight
-        self.edge_weights_ci = edge_weight_ci
         self.edge_weights_pop = edge_weight_pop
         self.node_weights = node_weight
-        self.node_weights_ci = node_weight_ci
         self.node_weights_pop = node_weight_pop
         self.edge_list = edge_list_out
         self.node_list = node_list_string
@@ -684,7 +578,7 @@ class Hypergraph(object):
             tolerance=1e-6,
             max_iterations=100,
             random_seed=12345,
-            bootstrap_samples=1,
+            normalisation="elements"
         ):
 
         """
@@ -739,20 +633,17 @@ class Hypergraph(object):
 
             random_seed : int, optional
                 The random seed to use in generating the initial vector (default: 12345)
-
-            bootstrap_samples : int, optional
-                The number of bootstrap samples to use to estimate the uncertainties in the
-                eigenvector centrality. (default: 1)
+            
+            normalisation : string, optional
+                The method to use for normalising the eigenvector. The options are 
+                "elements" - normalises by the theoretical maximum number of edges in the 
+                graph
+                "unit" - normalises the eigenvector so it has length one. 
+                (default: "elements")
 
         Returns
         -------
 
-            float
-                The calculated largest eigenvalue of the adjacency matrix
-
-            float
-                The calculated error in the eigenvalue
-
             numpy.array
                 the eigenvector centrality of the hypergraph. The order of the elements is
                 the same as the order passed in via the incidence matrix.
@@ -765,7 +656,7 @@ class Hypergraph(object):
         # Thank you Ed!
 
         # 0) setup
-        rng = np.random.default_rng(random_seed)
+        rng = random.default_rng(random_seed)
 
         if rep == "standard":
             inc_mat_original = self.incidence_matrix.T
@@ -775,6 +666,8 @@ class Hypergraph(object):
             weight_population = self.edge_weights_pop
             weight_resultant = self.node_weights
             resultant_pop = self.node_weights_pop
+            
+            n_nodes = len(self.node_list)
         elif rep == "dual":
             inc_mat_original = self.incidence_matrix
             old_eigenvector_estimate = rng.random(self.incidence_matrix.shape[0], dtype='float64')
@@ -783,34 +676,33 @@ class Hypergraph(object):
             weight_population = self.node_weights_pop
             weight_resultant = self.edge_weights
             resultant_pop = self.edge_weights_pop
+            
+            n_nodes = 2 ** len(self.node_list) - len(self.node_list) - 2
         elif rep == "bipartite":
 
             # We treat the bipartite representation differently -
             # the adjacency matrix is sparse, so we can use a sparse matrix
             # datatype and compute the eigevector directly
 
-            eigenvector_boot = []
-            eigenvalue_boot = []
+            inc_mat = self.incidence_matrix
+            weight = self.edge_weights
 
-            for _ in range(bootstrap_samples):
+            eig_val, eig_vec = _bipartite_eigenvector(
+                inc_mat,
+                weight
+            )
 
-                inc_mat = self.incidence_matrix
-                if bootstrap_samples > 1:
-                    weight = randomize_weights(self.edge_weights_pop.astype(np.int32), self.edge_weights)
-                else:
-                    weight = self.edge_weights
+            if normalisation == "elements":
+                norm_factor = sqrt(2 ** len(self.node_list) - 2)
+            elif normalisation == "unit":
+                norm_factor = norm(eig_vec)
+            else:
+                raise Exception("Normalisation method not supported.")
 
-                eig_val, eig_vec = _bipartite_eigenvector(
-                    inc_mat,
-                    weight
-                )
+            eigenvector = array(eig_vec / norm_factor)
+            eigenvalue = eig_val
 
-                eigenvector_boot.append(eig_vec / np.sum(eig_vec))
-                eigenvalue_boot.append(eig_val)
-
-            eigenvector_boot = np.array(eigenvector_boot)
-
-            return eigenvector_boot.mean(axis=0), eigenvector_boot.var(axis=0)
+            return eigenvector
 
         else:
             raise Exception("Representation not supported.")
@@ -830,92 +722,91 @@ class Hypergraph(object):
 
         # 2) do the Chebyshev
 
-        eigenvector_boot = []
-        eigenvalue_boot = []
+        if weighted_resultant:
+            res_weight = weight_resultant
+        else:
+            res_weight = ones_like(weight_resultant)
 
-        for _ in range(bootstrap_samples):
+        weight = weight_original
 
-            # apply a perturbation to the weights. Note, we will not do this if the
-            # user has requested only 1 bootstrap iteration or if the uncertainties in the
-            # weights are set to zero.
+        weight_norm = norm(weight)
+        res_weight_norm = norm(res_weight)
 
-            if bootstrap_samples > 1: # only perturb the weights if there is more than one sample
-                if weighted_resultant:
-                    res_weight = randomize_weights(resultant_pop.astype(np.int32), weight_resultant)
-                    inc_mat = ssp.diags(np.sqrt(res_weight)).dot(inc_mat_original)
-                else:
-                    inc_mat = inc_mat_original
+        # I'm not sure if we actually need to apply the normalisation to the vectors,
+        # since the eigenvector is being normalised in the end anyway. I don't think this
+        # is using that much time compared to the rest of the function so I will leave it
+        # in for now.
+        weight = weight / weight_norm
+        res_weight = res_weight / res_weight_norm
 
-                weight = randomize_weights(weight_population.astype(np.int32), weight_original)
-            else:
-                if weighted_resultant:
-                    inc_mat = ssp.diags(np.sqrt(weight_resultant)).dot(inc_mat_original)
-                else:
-                    inc_mat = inc_mat_original
-                weight = weight_original
+        inc_mat = diags(sqrt(res_weight)).dot(inc_mat_original)
+        old_eigenvector_estimate /= norm(old_eigenvector_estimate)
+        eigenvalue_estimates, eigenvalue_error_estimates = [], []
 
+        # In principle, the body of this loop could be compiled with Numba.
+        # However, iterate_vector() is quadratic in long_axis_size, whereas
+        # all the other operations here are linear in it, so we are spend very
+        # little time in the rest of this loop body.
+        for iteration in range(max_iterations):
 
-            old_eigenvector_estimate /= np.linalg.norm(old_eigenvector_estimate)
+            if self.verbose:
+                print("\rRunning iteration {}...".format(iteration), end="")
 
-            eigenvalue_estimates, eigenvalue_error_estimates = [], []
+            new_eigenvector_estimate = _iterate_vector(
+                inc_mat,
+                weight,
+                old_eigenvector_estimate
+            )
 
-            # In principle, the body of this loop could be compiled with Numba.
-            # However, iterate_vector() is quadratic in long_axis_size, whereas
-            # all the other operations here are linear in it, so we are spend very
-            # little time in the rest of this loop body.
-            for iteration in range(max_iterations):
+            # To estimate eigenvalue, take ratio of new to old eigenvector
+            # ignoring zeroes
+            mask = (new_eigenvector_estimate != 0) & (old_eigenvector_estimate != 0)
+            iter_eigenvalue_estimates = new_eigenvector_estimate[mask] / old_eigenvector_estimate[mask]
+            eigenvalue_estimate = iter_eigenvalue_estimates.mean()
+            eigenvalue_error_estimate = iter_eigenvalue_estimates.std()
+
+            eigenvalue_estimates.append(eigenvalue_estimate)
+            eigenvalue_error_estimates.append(eigenvalue_error_estimate)
+
+            if eigenvalue_error_estimate / eigenvalue_estimate < tolerance:
 
                 if self.verbose:
-                    print("\rRunning iteration {}...".format(iteration), end="")
-
-                new_eigenvector_estimate = _iterate_vector(
-                    inc_mat,
-                    weight,
-                    old_eigenvector_estimate
-                )
-
-                # To estimate eigenvalue, take ratio of new to old eigenvector
-                # ignoring zeroes
-                mask = (new_eigenvector_estimate != 0) & (old_eigenvector_estimate != 0)
-                iter_eigenvalue_estimates = new_eigenvector_estimate[mask] / old_eigenvector_estimate[mask]
-                eigenvalue_estimate = iter_eigenvalue_estimates.mean()
-                eigenvalue_error_estimate = iter_eigenvalue_estimates.std()
-
-                eigenvalue_estimates.append(eigenvalue_estimate)
-                eigenvalue_error_estimates.append(eigenvalue_error_estimate)
-
-                if eigenvalue_error_estimate / eigenvalue_estimate < tolerance:
-
-                    if self.verbose:
-                        print(
-                            "\nConverged at largest eigenvalue {:.2f} ± {:.4f} after {} iterations".format(
-                                eigenvalue_estimate,
-                                eigenvalue_error_estimate,
-                                iteration
-                            )
-                        )
-                    break
-
-                # Normalise to try to prevent overflows
-                old_eigenvector_estimate = (
-                    new_eigenvector_estimate /
-                    np.linalg.norm(new_eigenvector_estimate)
-                )
-
-            else:
-                if self.verbose:
-                    print("\nFailed to converge after", iteration, "iterations.")
-                    print("Last estimate was {:.2f} ± {:.4f}".format(
-                        eigenvalue_estimate,
-                        eigenvalue_error_estimate
+                    print(
+                        "\nConverged at largest eigenvalue {:.2f} ± {:.4f} after {} iterations".format(
+                            eigenvalue_estimate,
+                            eigenvalue_error_estimate,
+                            iteration
                         )
                     )
+                break
 
-            eigenvalue_boot.append(eigenvalue_estimate)
-            eigenvector_boot.append(new_eigenvector_estimate / np.sum(new_eigenvector_estimate))
+            # Normalise to try to prevent overflows
+            old_eigenvector_estimate = (
+                new_eigenvector_estimate /
+                norm(new_eigenvector_estimate)
+            )
 
-        eigenvector_boot = np.array(eigenvector_boot)
-        return eigenvector_boot.mean(axis=0), eigenvector_boot.var(axis=0)
+        else:
+            if self.verbose:
+                print("\nFailed to converge after", iteration, "iterations.")
+                print("Last estimate was {:.2f} ± {:.4f}".format(
+                    eigenvalue_estimate,
+                    eigenvalue_error_estimate
+                    )
+                )
+
+        eigenvalue = eigenvalue_estimate
+    
+        if normalisation == "elements":
+            norm_factor = sqrt(n_nodes)
+        elif normalisation == "unit":
+            norm_factor = norm(eig_vec)
+        else:
+            raise Exception("Normalisation method not supported.")        
+        
+        eigenvector = array(new_eigenvector_estimate / norm_factor)
+
+        return eigenvector
 
 
     def degree_centrality(
@@ -967,12 +858,12 @@ class Hypergraph(object):
         if rep == "standard":
             ax = 0
             if weighted:
-                M = ssp.diags(self.edge_weights).dot(self.incidence_matrix)
+                M = diags(self.edge_weights).dot(self.incidence_matrix)
 
         elif rep == "dual":
             ax = 1
             if weighted:
-                M = ssp.csr_matrix(self.incidence_matrix).dot(ssp.diags(self.node_weights))
+                M = csr_matrix(self.incidence_matrix).dot(diags(self.node_weights))
 
         else:
             raise Exception("Representation not supported.")
@@ -981,10 +872,75 @@ class Hypergraph(object):
 
 if __name__ == "__main__":
 
+
+    
+    @numba.jit(
+        nopython=True,
+        nogil=True,
+        fastmath=True,
+    )
+    def age_adjusted_weights(data, inds, *args):
+        """
+        This function returns a 1.0 divided by a number passed in as an optional
+        argument.
+        """
+        
+        age_weights = np.array( # European standard population 2013 weights 
+            [[0  , 4     , 0.050],
+             [5  , 9     , 0.055],
+             [10 , 14    , 0.055],
+             [15 , 19    , 0.055],
+             [20 , 24    , 0.060],
+             [25 , 29    , 0.060],
+             [30 , 34    , 0.065],
+             [35 , 39    , 0.070],
+             [40 , 44    , 0.070],
+             [45 , 49    , 0.070],
+             [50 , 54    , 0.070],
+             [55 , 59    , 0.065],
+             [60 , 64    , 0.060],
+             [65 , 69    , 0.055],
+             [70 , 74    , 0.050],
+             [75 , 79    , 0.040],
+             [80 , 84    , 0.025],
+             [85 , 90    , 0.015],
+             [90 , 94    , 0.008],
+             [95 , np.inf, 0.002]]
+        )
+        
+        
+        age_specific_numerator = np.zeros(age_weights.shape[0])
+        age_specific_denominator = np.zeros(age_weights.shape[0])
+        
+        for i in range(args[0].shape[0]):
+        
+            what_age_band = (age_weights[:, 0] <= args[0][i]) & (age_weights[:, 1] >= args[0][i])
+            
+            n_diseases = 0
+            for j in inds:
+                n_diseases += data[i, j]
+            
+            if n_diseases == len(inds):
+                age_specific_numerator[what_age_band] += args[1][i]
+                age_specific_denominator[what_age_band] += 1.0
+            
+        
+        out = 0.0
+        out_temp = 0.0
+        for i in range(age_weights.shape[0]):
+            if age_specific_denominator[i] > 0:
+                out += age_weights[i, 2] * age_specific_numerator[i] / age_specific_denominator[i]
+                out_temp += age_specific_numerator[i] / age_specific_denominator[i]
+        
+        return out, 0.0
+
+
+
     n_people = 5000
     n_diseases = 10
-
+    
     import pandas as pd
+    import numpy as np
 
     data = (np.random.rand(n_people, n_diseases) > 0.8).astype(np.uint8)
     data_pd = pd.DataFrame(
@@ -992,14 +948,33 @@ if __name__ == "__main__":
     ).rename(
         columns={i: "disease_{}".format(i) for i in range(data.shape[1])}
     )
+    
+    #data_age = np.random.randint(1, 100, n_people)
+    data_age = np.random.poisson(lam=50, size=n_people) # this is shit
+    data_HCRU = np.random.poisson(size=n_people)
+    
+    
+    data_array = data_pd.to_numpy().astype(uint8)
+    inds = np.array([0, 1])
+    
+    print(age_adjusted_weights(data_array, inds, data_age, data_HCRU))
+    
+    
 
-    h = Hypergraph(verbose=False)
-    h.compute_hypergraph(data_pd)
-
-    e_vec, e_vec_err = h.eigenvector_centrality(
-        weighted_resultant=True,
-        bootstrap_samples=1
+    h = Hypergraph(verbose=True)
+    h.compute_hypergraph(
+        data_pd,
+        age_adjusted_weights,
+        data_age, data_HCRU
     )
 
+    print(h.edge_weights)
 
+    e_vec, e_vec_err = h.eigenvector_centrality(
+        rep="dual",
+        weighted_resultant=True,
+    )
+    
+    print(e_vec)
 
+    
\ No newline at end of file
